@@ -1,14 +1,16 @@
+#![cfg_attr(feature = "benchmark", feature(test))]
+
 /// The module _seq_ provides a generic sequence container _Seq_ for Rust.
 ///
 /// _Seq_ is a lightweight container of data sequences (LIFO), managing dynamic list without
 /// dynamic memory allocation involved. Sequences are stored in stack frames of function contexts.
 /// Each element of a sequence has an individual lifetime `'a` managed by the Rust compiler.
 ///
-/// You can use Seq in your project adding the following dependency to your Cargo.toml file:
+/// Put this in your Cargo.toml:
 /// ```toml
 /// ## Cargo.toml file
 /// [dependencies]
-/// seq = "0.3.0"
+/// seq = "0.4"
 /// ```
 
 use std::fmt;
@@ -73,13 +75,35 @@ pub enum Seq<'a, T: 'a> {
     ConsOwn(T, Box<Seq<'a, T>>),
 }
 
+/// Seq method implementations
+impl<'a, T: 'a> Seq<'a, T> {
+    /// Returns a reference to the head-element
+    pub fn head (&'a self) -> Option<&'a T> {
+        match self {
+            &Seq::Empty => Option::None,
+            &Seq::ConsRef(ref ft1, _) => Option::Some(&*ft1),
+            &Seq::ConsOwn(ref ft1, _) => Option::Some(&*ft1),
+        }
+    }
+
+    /// Returns reference to the tail
+    pub fn tail (&'a self) -> Option<&'a Seq<T>> {
+        match self {
+            &Seq::Empty => Option::None,
+            &Seq::ConsRef(_, ref rt1) => Option::Some(*rt1),
+            &Seq::ConsOwn(_, ref rt1) => Option::Some(&**rt1),
+        }
+    }
+
+}
+
 /// The seqdef! macro defines a stack-allocated sequence variable for the speficied data list,
 /// the last data item in the list will be the top most in the sequence.
 ///
 /// Example 1) Creating a seq variable s where 2 is the top most data item
-/// `seqdef!(s; seq::empty() => 0, 1, 2);`
+/// `seqdef!(s; empty() => 0, 1, 2);`
 ///
-/// Example 2) Creating a seq variable t without explicit seq::empty(). Seq is identical to `s`.
+/// Example 2) Creating a seq variable t without explicit empty(). Seq is identical to `s`.
 /// `seqdef!(t; 0, 1, 2);`
 ///
 /// Example 3) Creating a seq variable u, using Seq `s` as tail of example 1.
@@ -419,6 +443,7 @@ mod tests {
 
         assert_eq!(&s4, &s4);
     }
+
     #[test]
     fn test_iter() {
         let s0: &Seq<u32> = empty();
@@ -484,4 +509,425 @@ mod tests {
         }
         assert_eq!(x.unwrap().into_iter().fold(0i32, ops::Add::add), 2000*42);
     }
+
+    #[test]
+    fn test_head_tail() {
+        let s: &Seq<u32> = empty();
+        let s = Seq::ConsRef(1u32, s);
+        let s = Seq::ConsRef(2u32, &s);
+        let s = Seq::ConsRef(3u32, &s);
+
+        let ft = s.head();
+        let rt = s.tail();
+
+        assert_eq!(ft.unwrap(), &3);
+        assert_eq!(rt.unwrap().head().unwrap(), &2);
+    }
+
 }
+
+
+#[cfg(all(feature = "benchmark", test))]
+mod benchmark {
+    extern crate test;
+
+    use super::Seq;
+    use super::empty;
+    use std::ops;
+    use std::vec;
+    use std::collections::LinkedList;
+
+    // Returns cumulation of  0, 0+1, 0+1+2, 0+1+2+3, ... 0+1+2+..+(N-1)
+    fn sum_of_sums(n:u32) -> u32
+    {
+        let cumulated = (n*(n+1)*((2*n)+1)/6) + ((n*(n+1))/2);
+        cumulated/2
+    }
+
+    // Recursive function, adding an element and cumulate the sums, until N-1 is reached.
+    fn recurs_stack_list(l: &mut LinkedList<u32>, cnt: u32, n:u32) -> u32 {
+        if cnt < n {
+            l.push_back(cnt);
+            let sum = l.iter().fold(0u32, ops::Add::add);
+            let r = sum + recurs_stack_list( l, cnt + 1, n);
+            l.pop_back();
+            r
+        } else {
+            0
+        }
+    }
+
+    #[bench]
+    fn bench_list_008(b: &mut test::Bencher) {
+        const N: u32 = 8;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_list_016(b: &mut test::Bencher) {
+        const N: u32 = 16;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_list_032(b: &mut test::Bencher) {
+        const N: u32 = 32;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_list_064(b: &mut test::Bencher) {
+        const N: u32 = 64;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_list_128(b: &mut test::Bencher) {
+        const N: u32 = 128;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_list_256(b: &mut test::Bencher) {
+        const N: u32 = 256;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_list_512(b: &mut test::Bencher) {
+        const N: u32 = 512;
+
+        b.iter(|| {
+            let mut l = LinkedList::new();
+
+            let sum = recurs_stack_list(&mut l, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    // Recursive function, adding an element and cumulate the sums, until N-1 is reached.
+    fn recurs_stack_vec(v: &mut vec::Vec<u32>, cnt: u32, n:u32) -> u32 {
+        if cnt < n {
+            v.push(cnt);
+            let sum = v.iter().fold(0u32, ops::Add::add);
+            let r = sum + recurs_stack_vec( v, cnt + 1, n);
+            v.truncate(cnt as usize);
+            r
+        } else {
+            0
+        }
+    }
+
+    #[bench]
+    fn bench_vec_008(b: &mut test::Bencher) {
+        const N: u32 = 8;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_vec_016(b: &mut test::Bencher) {
+        const N: u32 = 16;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_vec_032(b: &mut test::Bencher) {
+        const N: u32 = 32;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_vec_064(b: &mut test::Bencher) {
+        const N: u32 = 64;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_vec_128(b: &mut test::Bencher) {
+        const N: u32 = 128;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_vec_256(b: &mut test::Bencher) {
+        const N: u32 = 256;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_vec_512(b: &mut test::Bencher) {
+        const N: u32 = 512;
+
+        b.iter(|| {
+            let mut v = vec::Vec::new();
+            let sum = recurs_stack_vec(&mut v, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    // Recursive function, adding an element and cumulate the sums, until N-1 is reached.
+    fn recurs_stack_seq(s: &Seq<u32>, cnt: u32, n:u32) -> u32 {
+        if cnt < n {
+            let ext_s = Seq::ConsRef(cnt, s);
+
+            let sum = ext_s.into_iter().fold(0u32, ops::Add::add);
+            sum + recurs_stack_seq(&ext_s, cnt + 1, n)
+        } else {
+            0
+        }
+    }
+
+    #[bench]
+    fn bench_seq_008(b: &mut test::Bencher) {
+        const N: u32 = 8;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_seq_016(b: &mut test::Bencher) {
+        const N: u32 = 16;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_seq_032(b: &mut test::Bencher) {
+        const N: u32 = 32;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_seq_064(b: &mut test::Bencher) {
+        const N: u32 = 64;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_seq_128(b: &mut test::Bencher) {
+        const N: u32 = 128;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+
+    #[bench]
+    fn bench_seq_256(b: &mut test::Bencher) {
+        const N: u32 = 256;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_seq_512(b: &mut test::Bencher) {
+        const N: u32 = 512;
+
+        b.iter(|| {
+            let sum = recurs_stack_seq(empty(), 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    // Recursive function, adding an element and cumulate the sums, until N-1 is reached.
+    fn recurs_stack_array(a: &mut [u32], cnt: u32, n: u32) -> u32 {
+        if cnt < n {
+            a[cnt as usize] = cnt;
+
+            let sum = a[..(cnt+1) as usize].into_iter().fold(0u32, ops::Add::add);
+            sum + recurs_stack_array(a, cnt+1, n)
+        } else {
+            0
+        }
+    }
+
+    #[bench]
+    fn bench_array_008(b: &mut test::Bencher) {
+        const N: u32 = 8;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_array_016(b: &mut test::Bencher) {
+        const N: u32 = 16;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_array_032(b: &mut test::Bencher) {
+        const N: u32 = 32;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_array_064(b: &mut test::Bencher) {
+        const N: u32 = 64;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_array_128(b: &mut test::Bencher) {
+        const N: u32 = 128;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+    #[bench]
+    fn bench_array_256(b: &mut test::Bencher) {
+        const N: u32 = 256;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+    #[bench]
+    fn bench_array_512(b: &mut test::Bencher) {
+        const N: u32 = 512;
+        b.iter(|| {
+            let mut a: [u32; N as usize] = [0; N as usize];
+
+            let sum = recurs_stack_array(&mut a, 0, N);
+            assert_eq!(sum, sum_of_sums(N-1));
+            sum
+        });
+    }
+
+}
\ No newline at end of file
